/****************************************************/
/* File: parse.c                                    */
/* The parser implementation for the TINY compiler  */
/* Compiler Construction: Principles and Practice   */
/* Kenneth C. Louden                                */
/****************************************************/

#include "globals.h"
#include "util.h"
#include "scan.h"
#include "parse.h"

static TokenType token; /* holds current token */

/* function prototypes for recursive calls */
static TreeNode * func_sequence(void);//ɺ
static TreeNode* function(void);//
static TreeNode* block(void);//䲿
static TreeNode* param_seq(void);//Ĳ
static TreeNode* actual_param_seq(void);//ʱʵб
static TreeNode* return_type(void);//úֵ
static TreeNode* type_seq(void);//úı
static TreeNode* return_stmt(void);//úķֵ
static TreeNode * stmt_sequence(void);
static TreeNode * statement(void);
static TreeNode * if_stmt(void);
static TreeNode * repeat_stmt(void);
static TreeNode * assign_stmt(void);
static TreeNode * read_stmt(void);
static TreeNode * write_stmt(void);
static TreeNode * exp(void);
static TreeNode * simple_exp(void);
static TreeNode * term(void);
static TreeNode * factor(void);
static TreeNode * while_stmt(void);//whileĵݹʵ
static TreeNode* define_stmt(void);//ĵݹʵ
static TreeNode* ID_seq(void);//УŸ
static TreeNode* ary_stmt(void);//

static void syntaxError(char * message)
{ fprintf(listing,"\n>>> ");
  fprintf(listing,"Syntax error at line %d: %s",lineno,message);
  Error = TRUE;
}

static void match(TokenType expected)
{ if (token == expected) token = getToken();
  else {
    syntaxError("unexpected token in match -> ");
    printToken(token,tokenString);
    fprintf(listing,"      ");
  }
}

//ɶ
TreeNode* func_sequence(void)
{
    TreeNode* t = function();
    fprintf(listing, "first function generated\n");
    TreeNode* p = t;
    while (token != ENDFILE )
    {
        TreeNode* q;
        fprintf(listing, "match next function\n");
        q = function();
        if (q != NULL) {
            if (t == NULL) t = p = q;
            else /* now p cannot be NULL either */
            {
                p->sibling = q;
                fprintf(listing, "next function added to sibling\n");
                p = q;
            }
        }
    }
    return t;
}

/**/
TreeNode* function(void)
{
    TreeNode* t = NULL;
    TreeNode* p = return_type();
    fprintf(listing, "return_type defined\n");

    if (token == MAIN)
    {//main
        t = newFuncNode(MainK);
        t->attr.name = copyString(tokenString);
        match(MAIN);
        fprintf(listing, "token main matched\n");
    }

    else {
        t = newFuncNode(NormK);//
        fprintf(listing, " normal function node generated\n");
    }

    t->child[0] = p;
    fprintf(listing, "return_type added\n");

    //ƥ亯
    if (t != NULL && token == ID) {
        t->attr.name = copyString(tokenString);
        match(ID);
    }
   
    match(LPAREN);
    fprintf(listing, "LPRAEN matched\n");
    if (token == RPAREN) {
        //ûв
        //t->child[1] = NULL;
        match(RPAREN);
        fprintf(listing, "RPRAEN matched\n");
        t->child[1] = block();
        fprintf(listing, "function without param is added\n");
    }
    else {
        //в
        t->child[1] = param_seq();
        fprintf(listing, "param is added\n");
        match(RPAREN);
        t->child[2] = block();
        fprintf(listing, "function with params is added\n");
    }
    return t;
}


/*úֵtype*/
TreeNode* return_type(void)
{
    TreeNode* t = newTypeNode(Return);
    if (token != INT && token != FLOAT && token!=VOID) {
        syntaxError("Invalid return type");
    }
    t->attr.name = copyString(tokenString);
    match(token);
    return t;

}

/*int x, float y*/
TreeNode* type_seq(void)
{
    TreeNode* t = newStmtNode(TypeSeqK);
    switch (token)
    {
    case INT:
        t = newTypeNode(Integer);
        match(INT);
        break;
    case FLOAT:
        t = newTypeNode(Float);
        match(FLOAT);
        break;
    default:
        syntaxError("Invalid parameter type");
        break;
    }
    t->attr.name = copyString(tokenString);
    match(ID);
    TreeNode* p = t;
    while (token == COMMA) {
        match(COMMA);
        TreeNode* q;
        q = type_seq();
        if (q != NULL) {
            if (t == NULL) t = p = q;
            else {
                p->sibling = q;
                p = q;
            }
        }
    }
    return t;
}

/*param_seq-> type id [param_seq]*/
TreeNode* param_seq(void)
{
    TreeNode* t = newStmtNode(ParamK);
    t->child[0] = type_seq();
    return t;
}

/*ʵ func(a,10,b[]) ʵΪexp*/
TreeNode* actual_param_seq(void)
{
    TreeNode* t = newStmtNode(ParamK);
    TreeNode* e = exp();
    t->child[0] = e;
    TreeNode* p = e;
    while (token == COMMA) {
        match(COMMA);
        TreeNode* q;
        q = exp();
        if (q != NULL) {
            if (e == NULL) e = p = q;
            else {
                p->sibling = q;
                p = q;
            }
        }
    }
}

/*block->BEGIN stmt_seq END*/
TreeNode* block(void)
{
    match(BEGIN);

    TreeNode* t = stmt_sequence();
    fprintf(listing, "stmt_sequence matched!\n");
    match(END);
    //fprintf(listing, "block END matched!\n");
    return t;
}

/*return_stmt->RETURN exp*/
TreeNode* return_stmt(void)
{
    TreeNode* t = newStmtNode(ReturnK);
    match(RETURN);
    if (t != NULL) t->child[0] = exp();
    return t;
}



/*stmt_sequence -> statement {SEMI statement} */
TreeNode * stmt_sequence(void)
{ TreeNode * t = statement();
  TreeNode * p = t;
  while ((token!=ENDFILE) && (token!=END) &&
         (token!=ELSE) && (token!=UNTIL))
  { TreeNode * q;
    match(SEMI);
    q = statement();
    if (q!=NULL) {
      if (t==NULL) t = p = q;
      else /* now p cannot be NULL either */
      { p->sibling = q;
        p = q;
      }
    }
  }
  return t;
}

TreeNode * statement(void)
{ TreeNode * t = NULL;
  switch (token) {
    case BEGIN: t = block(); break;
    case RETURN: t = return_stmt(); break;
    case IF : t = if_stmt(); break;
    case REPEAT : t = repeat_stmt(); break;
    case ID : t = assign_stmt(); break;
    case READ : t = read_stmt(); break;
    case WRITE : t = write_stmt(); break;
        /*ӵstatment*/
    case WHILE: t = while_stmt(); break;
    case INT: 
    case VOID:
    case FLOAT:
        t = define_stmt(); break;
    default : syntaxError("unexpected token in statement -> ");
              printToken(token,tokenString);
              token = getToken();
              break;
  } /* end case */
  return t;
}

TreeNode * if_stmt(void)
{ TreeNode * t = newStmtNode(IfK);
  match(IF);
  if (t!=NULL) t->child[0] = exp();
  match(THEN);
  if (t!=NULL) t->child[1] = stmt_sequence();
  if (token==ELSE) {
    match(ELSE);
    if (t!=NULL) t->child[2] = stmt_sequence();
  }
  match(END);
  return t;
}

TreeNode * repeat_stmt(void)
{ TreeNode * t = newStmtNode(RepeatK);
  match(REPEAT);
  if (t!=NULL) t->child[0] = stmt_sequence();
  match(UNTIL);
  if (t!=NULL) t->child[1] = exp();
  return t;
}

/*while*/
TreeNode* while_stmt(void)
{
    TreeNode* t = newStmtNode(WhileK);
    match(WHILE);
    match(LPAREN);
    if (t != NULL) t->child[0] = exp();
    match(RPAREN);
    match(DO);
    if (t != NULL) t->child[1] = stmt_sequence();
    match(END);
    return t;
}

TreeNode * assign_stmt(void)
{ TreeNode * t = newStmtNode(AssignK);
  if ((t!=NULL) && (token==ID))
    t->attr.name = copyString(tokenString);
  match(ID);
  match(ASSIGN);
  if (t!=NULL) t->child[0] = exp();
  return t;
}

/*ı*/
TreeNode* define_stmt(void)
{
    TreeNode* t = newStmtNode(DefineK);
    switch (token)
    {
    case INT:   
        match(INT);
        break;
    case FLOAT:
        match(FLOAT);
        break;
        
    default:
        syntaxError("Invalid type!");
        break;
    }
    
    TreeNode* p = newExpNode(IdK);
    p->attr.name = copyString(tokenString);
    t->child[0] = p;
    match(ID);
    fprintf(listing, "ID matched in define_stmt!\n");
   
    if (t != NULL && token == EQ)
    {
        match(EQ);
        t->child[1] = exp(); // int x = 1;
        fprintf(listing, "Generate \n");
    }
    else if(t!=NULL && token==COMMA) 
    { 
        //match(COMMA);
        t->child[1] = ID_seq();//int x, y, z;
        fprintf(listing, "2\n");
    }
    else if (t != NULL && token == LBRA)
    {
        //match(LBRA);
        t->child[1] = ary_stmt();//int ary[5]; or  int a[2][3];
        fprintf(listing, "3\n");
    }
    else if (t != NULL && token == LPAREN)
    {
        match(LPAREN);
        //TreeNode* q = newFuncNode();
        //TreeNode* temp = p;
        //p = q;
        p->child[1] = actual_param_seq();//int func(int a, int b);
        fprintf(listing, "4\n");
    }
    fprintf(listing, "end\n");
   
    return t;
}

/*ID*/
TreeNode* ID_seq(void)
{
    //TreeNode* t = newSmtNode(IdSeqK);
    TreeNode* t = NULL;
    
    while (token == COMMA)
    {
        match(COMMA);
        TreeNode* q = newStmtNode(IdSeqK);
        TreeNode* p = factor();
        if (q!=NULL)
        {           
            q->child[0] = p;
            q->child[1] = ID_seq();
            t = q;
            
            //q->sibling = p;
            //p->sibling = factor();
            //p = q;
        }

    }
    return t;
}

/*½ʶ*/
TreeNode* ary_stmt()
{
    TreeNode* t = NULL;
    
    while (token == LBRA)
    {
        match(LBRA);
        TreeNode* q = newStmtNode(AryK);
        TreeNode* p = newExpNode(ConstK);
        p->attr.val = atoi(tokenString);
        
        match(NUM);
        fprintf(listing, "NUM matched!\n");
        match(RBRA);
        fprintf(listing, "RBRA matched!\n");
        if (q != NULL)
        {
            q->child[0] = p;
            fprintf(listing, "NUM added to the syntax tree!\n");
            fprintf(listing, "left tree: %d, now:%d\n", p->attr.val, q->child[0]->attr.val);
            q->child[1] = ary_stmt();
            fprintf(listing, "new ary_stmt added to the syntax tree!\n");
            t = q;
            fprintf(listing, "t's left tree: %d, now:%d\n", p->attr.val, t->child[0]->attr.val);
        }
        
    }

    return t;
}


TreeNode * read_stmt(void)
{ TreeNode * t = newStmtNode(ReadK);
  match(READ);
  if ((t!=NULL) && (token==ID))
    t->attr.name = copyString(tokenString);
  match(ID);
  return t;
}

TreeNode * write_stmt(void)
{ TreeNode * t = newStmtNode(WriteK);
  match(WRITE);
  if (t!=NULL) t->child[0] = exp();
  return t;
}


TreeNode * exp(void)
{ TreeNode * t = simple_exp();
  if ((token==LT)||(token==EQ)) {
    TreeNode * p = newExpNode(OpK);
    if (p!=NULL) {
      p->child[0] = t;
      p->attr.op = token;
      t = p;
    }
    match(token);
    if (t!=NULL)
      t->child[1] = simple_exp();
  }
  return t;
}

TreeNode * simple_exp(void)
{ TreeNode * t = term();
  while ((token==PLUS)||(token==MINUS))
  { TreeNode * p = newExpNode(OpK);
    if (p!=NULL) {
      p->child[0] = t;
      p->attr.op = token;
      t = p;
      match(token);
      t->child[1] = term();
    }
  }
  return t;
}



TreeNode * term(void)
{ TreeNode * t = factor();
  while ((token==TIMES)||(token==OVER))
  { TreeNode * p = newExpNode(OpK);
    if (p!=NULL) {
      p->child[0] = t;
      p->attr.op = token;
      t = p;
      match(token);
      p->child[1] = factor();
    }
  }
  return t;
}

TreeNode * factor(void)
{ TreeNode * t = NULL;
  char* temp;//id
  switch (token) {
    case NUM :
      t = newExpNode(ConstK);
      if ((t!=NULL) && (token==NUM))
        t->attr.val = atoi(tokenString);
      match(NUM);
      break;
    case ID :
      temp = copyString(tokenString);
      match(ID);
      /*ӶԪصĵ ID"["NUM"]"*/
      if (token == LBRA)
      {
          t = newExpNode(AryEleK);//Ԫؽڵ
          fprintf(listing, "AryEleK Node generated!\n");
          t->attr.name = copyString(temp);//
          match(LBRA);
          fprintf(listing, "AryEleK LBRA matched!\n");
          TreeNode* p = NULL;
          p = newExpNode(ConstK);
          if ((p != NULL) && (token == NUM))
          {
              p->attr.val = atoi(tokenString);
              //fprintf(listing, "index:%s\n", t->attr.val);
              t->child[0]= p;
          }
          match(NUM);
          fprintf(listing, "AryEleK index NUM matched!\n");
          match(RBRA);
          fprintf(listing, "AryEleK RBRA matched!\n");
      }
      /*Ժĵ*/
      else if (token == LPAREN) {
          t = newExpNode(FunCallK);
          t->attr.name = copyString(temp);
          match(LPAREN);
          t->child[0] = actual_param_seq();
          match(RPAREN);
      }
      else {
          t = newExpNode(IdK);
          t->attr.name = copyString(temp);
      }
      break;
    case LPAREN :
      match(LPAREN);
      t = exp();
      match(RPAREN);
      break;
    default:
      syntaxError("unexpected token in factor -> ");
      printToken(token,tokenString);
      token = getToken();
      break;
    }
  fprintf(listing, "factor generated!\n");
  return t;
}

/****************************************/
/* the primary function of the parser   */
/****************************************/
/* Function parse returns the newly 
 * constructed syntax tree
 */
TreeNode * parse(void)
{ TreeNode * t;
  token = getToken();
  //t = stmt_sequence();
  t = func_sequence();
  if (token!=ENDFILE)
    syntaxError("Code ends before file\n");
  return t;
}
